! =================================================================================================================================
! MODULE       : routing_tools
!
! CONTACT      : orchidee-help _at_ ipsl.jussieu.fr
!
! LICENCE      : IPSL (2016)
! This software is governed by the CeCILL licence see ORCHIDEE/ORCHIDEE_CeCILL.LIC
!
!>\BRIEF       
!!
!!\n DESCRIPTION: None
!!
!! RECENT CHANGE(S): None
!!
!! REFERENCE(S)	:
!!
!! SVN          :
!! $HeadURL: svn://forge.ipsl.jussieu.fr/orchidee/branches/ORCHIDEE-ROUTING/ORCHIDEE/src_sechiba/routing.f90 $
!! $Date: 2016-05-19 15:16:26 +0200 (Thu, 19 May 2016) $
!! $Revision: 3449 $
!! \n
!_ ================================================================================================================================
! 
MODULE routing_tools

  USE ioipsl   
  USE xios_orchidee
  USE ioipsl_para 
  USE constantes
  USE grid
  USE mod_orchidee_para

  IMPLICIT NONE
  PUBLIC :: routing_nextgrid_g, routing_anglediff_g
  ! We are working on the rectengular lat/lon grid of routing.nc
  ! The basic information of these grids are saved here.
  !
  INTEGER(i_std), PARAMETER        :: nbne=8
  ! The hypothesis of the flow directions used in routing.nc :
  REAL(r_std), PARAMETER, DIMENSION(nbne)     :: rose=(/0.0, 45.0, 90.0, 135.0, 180.0, 225.0, 270.0, 315.0/)
  !
CONTAINS
!! ================================================================================================================================
!! SUBROUTINE 	: routing_nextgrid_g
!!
!>\BRIEF         This function finds the grid box into which the flow direction points and we should
!!               be able to find the next sub-basin of the river.
!!               We will assume that this function is called on the root processor and with the global index of the
!!               grid point.
!!
!! DESCRIPTION (definitions, functional, design, flags) : None
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S):
!!
!! REFERENCES   : None
!!
!! FLOWCHART    : None
!! \n
!_ ================================================================================================================================
  !
INTEGER(i_std) FUNCTION routing_nextgrid_g(pnt, trip, inc)
  !
  ! Input
  !
  INTEGER(i_std) :: pnt, trip
  INTEGER(i_std), OPTIONAL :: inc
  !
  ! Local
  !
  INTEGER(i_std)                   :: i
  REAL(r_std), DIMENSION(NbNeighb) :: diffangle
  INTEGER(i_std), DIMENSION(1)     :: imin
  INTEGER(i_std)                   :: trip_loc, nbind
  !
  ! Compute the angle between the direction of the routing.nc flow and the headings perpendicular of the segments
  ! of the polygon.
  !
  IF (.NOT. is_root_prc) THEN
     WRITE(*,*) "The function routing_nextgrid_g is written to work on the root proc and the full grid"
     STOP "routing_nextgrid_g"
  ENDIF
  !
  ! Make sure that the flow direction coming from routing.f90 is between 1 and 8. This allows to call the
  ! function without having to make sure trip is between 1 and 8.
  !
  trip_loc = MODULO(trip-1, nbne)+1
  !
  DO i=1,NbNeighb
     diffangle(i) = ABS(haversine_diffangle(headings_g(pnt,i), rose(trip_loc)))
  ENDDO
  !
  imin = MINLOC(diffangle)
  !
  IF ( PRESENT(inc) ) THEN
     ! Ensure that the increment the users asks is within the NbNeighb the polygone has.
     nbind = MODULO(imin(1)+inc-1, NbNeighb)+1
  ELSE
     nbind = imin(1)
  ENDIF
  !
  routing_nextgrid_g = neighbours_g(pnt,nbind)
  !
END FUNCTION routing_nextgrid_g
!
!! ================================================================================================================================
!! SUBROUTINE 	: routing_nextgrid
!!
!>\BRIEF         This function finds the grid box into which the flow direction points and we should
!!               be able to find the next sub-basin of the river.
!!               We will assume that this function is called on the root processor and with the global index of the
!!               grid point. This verison of the function can be called in a parallel region.
!!
!! DESCRIPTION (definitions, functional, design, flags) : None
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S):
!!
!! REFERENCES   : None
!!
!! FLOWCHART    : None
!! \n
!_ ================================================================================================================================
  !
INTEGER(i_std) FUNCTION routing_nextgrid(pnt, trip, inc)
  !
  ! Input
  !
  INTEGER(i_std) :: pnt, trip
  INTEGER(i_std), OPTIONAL :: inc
  !
  ! Local
  !
  INTEGER(i_std)                   :: i
  REAL(r_std), DIMENSION(NbNeighb) :: diffangle
  INTEGER(i_std), DIMENSION(1)     :: imin
  INTEGER(i_std)                   :: trip_loc, nbind
  !
  ! Make sure that the flow direction coming from routing.f90 is between 1 and 8. This allows to call the
  ! function without having to make sure trip is between 1 and 8.
  !
  trip_loc = MODULO(trip-1, nbne)+1
  !
  ! Compute the angle between the direction of the routing.nc flow and the headings perpendicular of the segments
  ! of the polygon.
  !
  DO i=1,NbNeighb
     diffangle(i) = ABS(haversine_diffangle(headings(pnt,i), rose(trip_loc)))
  ENDDO
  !
  imin = MINLOC(diffangle)
  !
  IF ( PRESENT(inc) ) THEN
     ! Ensure that the increment the users asks is within the NbNeighb the polygone has.
     nbind = MODULO(imin(1)+inc-1, NbNeighb)+1
  ELSE
     nbind = imin(1)
  ENDIF
  !
  routing_nextgrid = neighbours(pnt,nbind)
  !
END FUNCTION routing_nextgrid
!
!
!! ================================================================================================================================
!! SUBROUTINE 	: routing_anglediff_g
!!
!>\BRIEF         This routine computes the angle difference between the flow direction on the routing.nc grid (ytip) and 
!!               the heading of the line which links the 2 points (pnt and pntout) of the model grid. These 2 grid
!!               points have to be neighbours.
!!
!! DESCRIPTION (definitions, functional, design, flags) : None
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S):
!!
!! REFERENCES   : None
!!
!! FLOWCHART    : None
!! \n
!_ ================================================================================================================================
  !
REAL(r_std) FUNCTION routing_anglediff_g(pnt, pntout, trip)
  !
  ! Input
  !
  INTEGER(i_std) :: pnt, pntout
  INTEGER(i_std) :: trip
  !
  ! Local
  !
  INTEGER(i_std)                      :: i
  INTEGER(i_std)                      :: imin
  INTEGER(i_std)                      :: trip_loc
  !
  ! This version of the routine can only be called on the root proc.
  ! 
  IF (.NOT. is_root_prc) THEN
     WRITE(*,*) "The function routing_flowangle_g is written to work on the root proc and the full grid"
     STOP "routing_flowangle_g"
  ENDIF
  trip_loc = MOD(trip+nbne-1, nbne)+1
  !
  ! Find the index of the neighbour we are considering so we get the heading out of pnt
  !
  imin=-1
  DO i=1,NbNeighb
     IF ( neighbours_g(pnt,i) .EQ. pntout ) THEN
        imin = i
     ENDIF
  ENDDO
  IF ( i > 0 ) THEN
     routing_anglediff_g = ABS(haversine_diffangle(headings_g(pnt,imin), rose(trip_loc)))
  ELSE
     WRITE(*,*) "Input arguments : pnt, pntout, trip = ", pnt, pntout, trip
     CALL ipslerr_p(3,'routing_anglediff_g', &
                    "Cannot find the neighbour requested.", &
             &      "None of the neighbours of the specified point (pnt) match pntout.", "")
  ENDIF
  !
END FUNCTION routing_anglediff_g
!
!
!! ================================================================================================================================
!! SUBROUTINE 	: routing_anglediff
!!
!>\BRIEF         This routine computes the angle difference between the flow direction on the routing.nc grid (ytip) and 
!!               the heading of the line which links the 2 points (pnt and pntout) of the model grid. These 2 grid
!!               points have to be neighbours.
!!
!! DESCRIPTION (definitions, functional, design, flags) : None
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S):
!!
!! REFERENCES   : None
!!
!! FLOWCHART    : None
!! \n
!_ ================================================================================================================================
  !
REAL(r_std) FUNCTION routing_anglediff(pnt, pntout, trip)
  !
  ! Input
  !
  INTEGER(i_std) :: pnt, pntout
  INTEGER(i_std) :: trip
  !
  ! Local
  !
  INTEGER(i_std)    :: i
  INTEGER(i_std)    :: imin
  INTEGER(i_std)    :: trip_loc
  !
  ! This version of the routine can only be called on the root proc.
  ! 
  trip_loc = MOD(trip+nbne-1, nbne)+1
  !
  ! Find the index of the neighbour we are considering so we get the heading out of pnt
  !
  imin=-1
  DO i=1,NbNeighb
     IF ( neighbours(pnt,i) .EQ. pntout ) THEN
        imin = i
     ENDIF
  ENDDO
  IF ( i > 0 ) THEN
     routing_anglediff = ABS(haversine_diffangle(headings(pnt,imin), rose(trip_loc)))
  ELSE
     WRITE(*,*) "Input arguments : pnt, pntout, trip = ", pnt, pntout, trip
     CALL ipslerr_p(3,'routing_anglediff', &
                    "Cannot find the neighbour requested.", &
             &      "None of the neighbours of the specified point (pnt) match pntout.", "")
  ENDIF
  !
END FUNCTION routing_anglediff
!
!
!!
!! ================================================================================================================================
!! SUBROUTINE 	: routing_bestsubbasin()
!!
!>\BRIEF         This routine help routing_linkup to find the most suitable sub-basin in the grid box
!!               in which the source sub-basin flows.
!!
!! DESCRIPTION (definitions, functional, design, flags) : None
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S):
!!
!! REFERENCES   : None
!!
!! FLOWCHART    : None
!! \n
!_ ================================================================================================================================
!
SUBROUTINE routing_bestsubbasin(src_grid, src_id, src_hierarchy, src_flowdir, invented_basins, &
     &                          ptmax, basmax, pt, nbsubbas, ids, hierarchy, flowdir, nbcoastals, coastals, &
     &                          outbas, quality)
  !
  ! Input
  !
  INTEGER(i_std), INTENT(in) :: src_grid, src_id, src_flowdir
  REAL(r_std), INTENT(in)    :: src_hierarchy
  REAL(r_std), INTENT(in)    :: invented_basins
  !
  INTEGER(i_std), INTENT(in) :: ptmax, basmax, pt 
  INTEGER(i_std), INTENT(in) :: nbsubbas(ptmax)
  INTEGER(i_std), INTENT(in) :: ids(ptmax,basmax), flowdir(ptmax,basmax)
  REAL(r_std), INTENT(in)    :: hierarchy(ptmax,basmax)
  INTEGER(i_std), INTENT(in) :: nbcoastals(ptmax)
  INTEGER(i_std), INTENT(in) :: coastals(ptmax,basmax)
  !
  ! Output
  !
  INTEGER(i_std), INTENT(out) :: outbas
  REAL(r_std), INTENT(out)    :: quality
  !
  ! Local
  !
  INTEGER(i_std) :: i, sbl, nbfbas, nbopt
  INTEGER(i_std), DIMENSION(1) :: ff
  INTEGER(i_std), DIMENSION(basmax) :: fbas, tbas, sboptions, angle, subbas
  INTEGER(i_std), DIMENSION(basmax) :: fbas_flowdir, tbas_flowdir
  REAL(r_std), DIMENSION(basmax) :: fbas_hierarchy, tbas_hierarchy
  !
  ! Set some default values
  !
  nbfbas = 0
  nbopt = 0
  sboptions(:) = undef_int
  outbas = undef_int
  !
  ! 1.0 Find possible basins in the list of the target grid 
  !
  DO sbl=1,nbsubbas(pt)
     !
     ! Either it is a standard basin or one aggregated from ocean flow points.
     ! If we flow into a another grid box nbsubbas(pt)we have to make sure that its hierarchy in the
     ! basin is lower.
     !
     IF ( ids(pt,sbl) .GT. 0 ) THEN
        IF ( ids(pt,sbl) .EQ. src_id .OR. ids(pt,sbl) .GT. invented_basins) THEN
           nbfbas =nbfbas + 1
           tbas(nbfbas) = sbl
           tbas_hierarchy(nbfbas) = hierarchy(pt,sbl)
           tbas_flowdir(nbfbas) = flowdir(pt,sbl)
        ENDIF
     ELSE
        IF ( COUNT(coastals(pt,1:nbcoastals(pt)) .EQ. src_id) .GT. 0 ) THEN
           nbfbas =nbfbas + 1
           tbas(nbfbas) = sbl
           tbas_hierarchy(nbfbas) = hierarchy(pt,sbl)
        ENDIF
     ENDIF
  ENDDO
  !
  ! Sort the possible basins by hierarchy so we can find the best placed sub-basin.
  !
  IF (nbfbas .GT. 1) THEN
     !
     IF ( MAXVAL(hierarchy(pt,1:nbsubbas(pt))) >= undef_int ) THEN
        CALL ipslerr_p(3,'routing_bestsubbasin', &
             "Cannot sort the sub-basins by hierarchy.", &
             &           "Some alternate method needs to be found or undef_int increased.", "")
     ENDIF
     !
     DO i=1,nbfbas
        ff = MINLOC(tbas_hierarchy(1:nbfbas))
        fbas(i) = tbas(ff(1))
        fbas_hierarchy(i) = tbas_hierarchy(ff(1))
        fbas_flowdir(i) = tbas_flowdir(ff(1))
        tbas_hierarchy(ff(1)) = undef_int
     ENDDO
  ELSE
     fbas(:) = tbas(:)
     fbas_hierarchy(:) = tbas_hierarchy(:)
     fbas_flowdir(:) = tbas_flowdir(:)
  ENDIF
  !
  ! Analyse our options if there are at least one
  !
  sbl = undef_int
  IF (nbfbas .EQ. 1) THEN
     !
     ! We check that this only option is valid to provide a quality flag.
     ! If hierachy is lower then we have no problems.
     ! If the hierrachies are equal then it is possible if both sub-basins 
     ! flow into the same direction.
     !
     IF (  fbas_hierarchy(1) < src_hierarchy ) THEN
        sbl = fbas(1)
     ELSE IF ( (fbas_hierarchy(1) - src_hierarchy ) < EPSILON(src_hierarchy) ) THEN
        IF ( fbas_flowdir(1) == src_flowdir ) THEN
           sbl = fbas(1)
        ELSE
           sbl = undef_int
        ENDIF
     ELSE
        sbl = undef_int
     ENDIF
     !
  ELSE IF (nbfbas .GT. 1) THEN
     !
     ! First work on the hierarchy based on the sorting done above.
     !
     ff = MINLOC(ABS(fbas_hierarchy(1:nbfbas)-src_hierarchy))
     !
     ! Get the basin with the hierarchy just below src_hierarchy
     !
     IF ( fbas_hierarchy(ff(1)) < src_hierarchy ) THEN
        sbl = fbas(ff(1))
     ELSE
        !
        !
        IF ( ff(1) > 1 ) THEN
           ! If we are in the case of fbas_hierarchy(ff(1)) = src_hierarch we know we have a point with lower
           ! hierarchy so we flow there.
           sbl = fbas(ff(1)-1)
           nbopt=0
        ELSE
           ! We are probably in the fbas_hierarchy(ff(1)) = src_hierarch situation and we need to know if we have 
           ! more sub-basins in this case
           IF ( ABS(fbas_hierarchy(ff(1))-src_hierarchy) < EPSILON(src_hierarchy) ) THEN
              sbl = fbas(ff(1))
              nbopt=0
              DO i=1,nbfbas
                 IF ( ABS(fbas_hierarchy(i)-src_hierarchy) < EPSILON(src_hierarchy) ) THEN
                    nbopt = nbopt+1
                    sboptions(nbopt) = fbas(i)
                 ENDIF
              ENDDO
           ELSE
!!$              CALL ipslerr_p(2,'routing_bestsubbasin', &
!!$                   &           "The sub-basins with the closest hierrachy has still a higher value.",&
!!$                   &           "This is not an option and we will try elsewhere.", "")
              sbl = undef_int
           ENDIF
        ENDIF

     ENDIF
  ELSE
     sbl = undef_int
  ENDIF
  !
  ! Now we have either found a basin or we have a number of options we can explore
  !
  IF ( sbl == undef_int ) THEN
     outbas = undef_int 
     quality = 0
  ELSE
     IF ( nbopt == 0 ) THEN
        outbas = sbl
        quality = 1
     ELSE
        !
        ! Go through the list of options and try to find a solution
        !
        ! a) Similar flow directions. Here we are working with the integer values provided on the 
        ! regular lat/lon grid. Thus we always have 8 neighbours.
        !
        angle(:) = undef_int
        subbas(:) = undef_int
        DO i=1,nbopt
           angle(i) = MIN(MOD(src_flowdir+8 - flowdir(pt,sboptions(i)), 8), &
                &         MOD(flowdir(pt,sboptions(i))+8 - src_flowdir, 8))
           subbas(i) = sboptions(i)
        ENDDO
        ff=MINLOC(angle)
        IF (angle(ff(1)) <= 1) THEN
           outbas = subbas(ff(1))
           quality = 0.5
        ENDIF
        !
        ! b) Is the next outflow options is the ocean, then it is OK as well and even a little better.
        !
        DO i=1,nbopt
           IF ( flowdir(pt,sboptions(i)) < 0 ) THEN
              outbas = sboptions(i)
              quality = 0.75
           ENDIF
        ENDDO
     ENDIF
  ENDIF
  !
END SUBROUTINE routing_bestsubbasin
!
!
!!
!! ================================================================================================================================
!! SUBROUTINE 	: routing_reg_bestsubbasin()
!!
!>\BRIEF         This routine help routing_linkup to find the most suitable sub-basin in the grid box
!!               in which the source sub-basin flows.
!!
!! DESCRIPTION (definitions, functional, design, flags) : None
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S):
!!
!! REFERENCES   : None
!!
!! FLOWCHART    : None
!! \n
!_ ================================================================================================================================
!
SUBROUTINE routing_reg_bestsubbasin(src_grid, src_basin, src_id, src_hierarchy, src_fac, src_flowdir, invented_basins, &
     &                          ptmax, basmax, pt, nbsubbas, ids, hierarchy, fac, flowdir, nbcoastals, coastals, &
     &                          outbas, quality)
  !
  ! Input
  !
  INTEGER(i_std), INTENT(in) :: src_grid, src_id, src_flowdir, src_basin
  REAL(r_std), INTENT(in)    :: src_hierarchy, src_fac ! src_fac should be integer but nevermind
  REAL(r_std), INTENT(in)    :: invented_basins
  !
  INTEGER(i_std), INTENT(in) :: ptmax, basmax, pt 
  INTEGER(i_std), INTENT(in) :: nbsubbas(ptmax)
  INTEGER(i_std), INTENT(in) :: ids(ptmax,basmax), flowdir(ptmax,basmax)
  REAL(r_std), INTENT(in)    :: hierarchy(ptmax,basmax)
  REAL(r_std), INTENT(in)    :: fac(ptmax,basmax) ! fac should be integer but real is still ok
  INTEGER(i_std), INTENT(in) :: nbcoastals(ptmax)
  INTEGER(i_std), INTENT(in) :: coastals(ptmax,basmax)
  !
  ! Output
  !
  INTEGER(i_std), INTENT(out) :: outbas
  REAL(r_std), INTENT(out)    :: quality
  !
  ! Local
  !
  INTEGER(i_std) :: i, j, sbl, nbfbas, nbopt
  INTEGER(i_std), DIMENSION(1) :: ff
  INTEGER(i_std), DIMENSION(basmax) :: fbas, tbas, sboptions, angle, subbas
  INTEGER(i_std), DIMENSION(basmax) :: fbas_flowdir, tbas_flowdir
  REAL(r_std), DIMENSION(basmax) :: fbas_hierarchy, tbas_hierarchy
  REAL(r_std), DIMENSION(basmax) :: fbas_fac, tbas_fac
  !
  ! Debug
  !
  LOGICAL, PARAMETER :: debug = .FALSE. !! Logical variable to debug
  INTEGER(i_std)     :: testbasinid
  REAL(r_std)        :: maxhiediff
  !
  ! Set some default values
  !
  fbas = 0
  nbfbas = 0
  nbopt = 0
  sboptions(:) = undef_int
  outbas = undef_int
  testbasinid = 107448 
  !
  ! 1.0 Find possible basins in the list of the target grid 
  !
  DO sbl=1,nbsubbas(pt)
     !
     ! Either it is a standard basin or one aggregated from ocean flow points.
     ! If we flow into a another grid box nbsubbas(pt) we have to make sure that its hierarchy in the
     ! basin is lower.
     ! Trung: we get also information about flow accumulation (fac)
     !
     IF ( ids(pt,sbl) .GT. 0 ) THEN
        IF ( ids(pt,sbl) .EQ. src_id .OR. ids(pt,sbl) .GT. invented_basins) THEN
           nbfbas =nbfbas + 1
           tbas(nbfbas) = sbl
           tbas_hierarchy(nbfbas) = hierarchy(pt,sbl)
           tbas_fac(nbfbas) = fac(pt,sbl)
           tbas_flowdir(nbfbas) = flowdir(pt,sbl)
        ENDIF
     ELSE
        IF ( COUNT(coastals(pt,1:nbcoastals(pt)) .EQ. src_id) .GT. 0 ) THEN
           nbfbas =nbfbas + 1
           tbas(nbfbas) = sbl
           tbas_hierarchy(nbfbas) = hierarchy(pt,sbl)
           tbas_fac(nbfbas) = fac(pt,sbl)
        ENDIF
     ENDIF
  ENDDO
  !
  ! Sort the possible basins by hierarchy so we can find the best placed sub-basin.
  !
  IF (nbfbas .GT. 1) THEN
     !
     IF ( MAXVAL(hierarchy(pt,1:nbsubbas(pt))) >= undef_int ) THEN
        CALL ipslerr_p(3,'routing_reg_bestsubbasin', &
             "Cannot sort the sub-basins by hierarchy.", &
             &           "Some alternate method needs to be found or undef_int increased.", "")
     ENDIF
     !
     ! Trung: Using "hierarchy" as more important than fac but still need using
     ! "fac" to deal with "tiny" sub-basins which can twist the stream too much
     !
     DO i=1,nbfbas
        ff = MINLOC(tbas_hierarchy(1:nbfbas))
        fbas(i) = tbas(ff(1))
        fbas_hierarchy(i) = tbas_hierarchy(ff(1))
        fbas_fac(i) = tbas_fac(ff(1))
        fbas_flowdir(i) = tbas_flowdir(ff(1))
        tbas_hierarchy(ff(1)) = undef_int
     ENDDO
  ELSE
     fbas(:) = tbas(:)
     fbas_hierarchy(:) = tbas_hierarchy(:)
     fbas_fac(:) = tbas_fac(:)
     fbas_flowdir(:) = tbas_flowdir(:)
  ENDIF
  !
  ! Debug if you want
  !
  IF ( debug .AND. testbasinid > 0 ) THEN
     IF ( src_id .EQ. testbasinid )  THEN
        WRITE (numout,*) " = = = = = = S T A R T D E B U G = = = = = = "
        WRITE (numout,*) " Bestsubbasin TEST : ", src_id, "@", src_grid, src_basin
        WRITE (numout,*) " nbfbas = ", nbfbas, "@", pt
        WRITE (numout,*) " src_hierarchy = ", src_hierarchy
        WRITE (numout,*) " src_fac = ", src_fac
        WRITE (numout,*) " = = = = = = = = = = = = = = = = = = = = = = "
        IF ( nbfbas .GT. 0 ) THEN 
           WRITE (numout,*) " i, fbas(i), fbas_hierarchy(i), fbas_fac(i)" !, fbas_flowdir(i)"
           DO i=1,nbfbas
              WRITE (numout,*) i, fbas(i), fbas_hierarchy(i), fbas_fac(i) !, fbas_flowdir(i)
           ENDDO
        ELSE
           WRITE (numout,*) " We did not find any suitable sub-basin !"
        ENDIF
        WRITE (numout,*) " = = = = = = = E N D D E B U G = = = = = = = "
     ENDIF
  ENDIF
  !
  ! Analyse our options if there are at least one
  !
  sbl = undef_int
  IF (nbfbas .EQ. 1) THEN
     !
     ! We check that this only option is valid to provide a quality flag.
     ! If hierachy is lower then we have no problems.
     ! If the hierrachies are equal then it is possible if both sub-basins 
     ! flow into the same direction.
     !
     ! Trung: If we found only one suitable sub-basin, we don't care about "fac".
     !
     IF (  fbas_hierarchy(1) < src_hierarchy ) THEN
        sbl = fbas(1)
     ELSE IF ( (fbas_hierarchy(1) - src_hierarchy ) < EPSILON(src_hierarchy) ) THEN
        IF ( fbas_flowdir(1) == src_flowdir ) THEN
           sbl = fbas(1)
        ELSE
           sbl = undef_int
        ENDIF
     ELSE
        sbl = undef_int
     ENDIF
     !
  ELSE IF (nbfbas .GT. 1) THEN
     !
     ! First work on the hierarchy based on the sorting done above.
     !
     ! Trung: Instead of using MINLOC
     !        ff = MINLOC(ABS(fbas_hierarchy(1:nbfbas)-src_hierarchy))
     !        We need a loop here: There will be situations in which
     !        there are lots of sub-basin have same ZERO hierarchy.
     !        In this case, using MINLOC will return ff(1) = 1
     !        While, in fact, it should be greater than 1.
     !
     ff(1) = 1
     maxhiediff = ABS(fbas_hierarchy(ff(1))-src_hierarchy)
     ! Trung: Shorten iteration here by DO WHILE (instead of DO i = 1,nbfbas)
     j = 0
     i = ff(1)
     DO WHILE ( i .GE. 1 .AND. i .LT. nbfbas .AND. j .EQ. 0 )
        i = i + 1
        IF ( ABS(fbas_hierarchy(i)-src_hierarchy) .LE. maxhiediff ) THEN
           ff(1) = i
           maxhiediff = ABS(fbas_hierarchy(ff(1))-src_hierarchy)
        ELSE
           j = -1
        ENDIF
     ENDDO
     !
     ! Get the basin with the hierarchy just below src_hierarchy
     !
     ! Trung: put requirement about "fac" here
     !        ONLY coastal point can have fbas_fac(ff(1)) = src_fac, right?
     !        I need to think more carefully about it. 
     !        Now, I consider that there is NO fbas_fac(ff(1)) == src_fac
     !
     IF ( (fbas_hierarchy(ff(1)) .LT. src_hierarchy) .AND. (fbas_fac(ff(1)) .GT. src_fac) ) THEN
        sbl = fbas(ff(1))
     ELSE
        !
        !
        IF ( ff(1) > 1 ) THEN
           !
           ! If we are in the case of fbas_hierarchy(ff(1)) = src_hierarch we know we have a point with lower
           ! hierarchy so we flow there.
           !
           ! Trung: We consider "fbas_fac(ff(1)) > src_fac", so we need to go
           !        higher in sorted list fbas(:) to find suitable sub-basin.
           !
           j = 0
           i = ff(1)
           DO WHILE ( i .GT. 1 .AND. j .EQ. 0 )
              i = i - 1
              IF ( fbas_hierarchy(i) .LT. src_hierarchy .AND. fbas_fac(i) .GT. src_fac ) THEN
                 sbl = fbas(i)
                 nbopt = 0
                 j = -1
              ENDIF 
           ENDDO
        ELSE
           ! We are probably in the fbas_hierarchy(ff(1)) = src_hierarch situation and we need to know if we have 
           ! more sub-basins in this case
           ! Trung: If the quality of routing.nc is good, we will never come to
           !        this step, right?
           IF ( ABS(fbas_hierarchy(ff(1))-src_hierarchy) < EPSILON(src_hierarchy) ) THEN
              sbl = fbas(ff(1))
              nbopt=0
              DO i=1,nbfbas
                 ! Trung: the condition about "fac" here need to be investigated later
                 IF ( (ABS(fbas_hierarchy(i)-src_hierarchy) < EPSILON(src_hierarchy)) .AND. (fbas_fac(i) > src_fac) ) THEN
                    nbopt = nbopt+1
                    sboptions(nbopt) = fbas(i)
                 ENDIF
              ENDDO
           ELSE
!!$              CALL ipslerr_p(2,'routing_reg_bestsubbasin', &
!!$                   &           "The sub-basins with the closest hierrachy has still a higher value.",&
!!$                   &           "This is not an option and we will try elsewhere.", "")
              sbl = undef_int
           ENDIF
        ENDIF
     ENDIF
  ENDIF
  !
  ! Now we have either found a basin or we have a number of options we can explore
  !
  IF ( sbl == undef_int ) THEN
     outbas = undef_int 
     quality = 0
  ELSE
     IF ( nbopt == 0 ) THEN
        outbas = sbl
        quality = 1
     ELSE
        !
        ! Go through the list of options and try to find a solution
        !
        ! a) Similar flow directions. Here we are working with the integer values provided on the 
        ! regular lat/lon grid. Thus we always have 8 neighbours.
        !
        angle(:) = undef_int
        subbas(:) = undef_int
        DO i=1,nbopt
           angle(i) = MIN(MOD(src_flowdir+8 - flowdir(pt,sboptions(i)), 8), &
                &         MOD(flowdir(pt,sboptions(i))+8 - src_flowdir, 8))
           subbas(i) = sboptions(i)
        ENDDO
        ff=MINLOC(angle)
        IF (angle(ff(1)) <= 1) THEN
           outbas = subbas(ff(1))
           quality = 0.5
        ENDIF
        !
        ! b) Is the next outflow options is the ocean, then it is OK as well and even a little better.
        !
        DO i=1,nbopt
           IF ( flowdir(pt,sboptions(i)) < 0 ) THEN
              outbas = sboptions(i)
              quality = 0.75
           ENDIF
        ENDDO
     ENDIF
  ENDIF
  !
END SUBROUTINE routing_reg_bestsubbasin
!! ================================================================================================================================
!! SUBROUTINE 	: routing_updateflow
!!
!>\BRIEF         This subroutine allows to update the various variables which store the flow
!!               from one gridbox to another. It is just a tool to simplify the main code.
!!
!! DESCRIPTION (definitions, functional, design, flags) : None
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S):
!!
!! REFERENCES   : None
!!
!! FLOWCHART    : None
!! \n
!_ ================================================================================================================================
  !
SUBROUTINE routing_updateflow(sp, sb, ep, eb, pt, bas, basmax, out_grid, out_basin, in_number, in_grid, in_basin)
  !
  ! ARGUMENTS
  !
  INTEGER(i_std), INTENT(in) :: sp, sb          !! Source point and source basin
  INTEGER(i_std), INTENT(in) :: ep, eb          !! target point and target basin
  INTEGER(i_std), INTENT(in) :: pt, bas, basmax !! dimensions of arrays
  !
  INTEGER(i_std), DIMENSION(pt,bas), INTENT(inout)           :: out_grid  !! Grid to which we flow
  INTEGER(i_std), DIMENSION(pt,bas), INTENT(inout)           :: out_basin !! Basin in the out_grid
  INTEGER(i_std), DIMENSION(pt,basmax), INTENT(inout)        :: in_number !! Number of basins flowing into this grid
  INTEGER(i_std), DIMENSION(pt,basmax,basmax), INTENT(inout) :: in_grid   !! Source grid for this grid
  INTEGER(i_std), DIMENSION(pt,basmax,basmax), INTENT(inout) :: in_basin  !! Source basin for this grid
  !
  ! LOCAL
  !
  INTEGER(i_std) :: i
  LOGICAL        :: transfer
  !
  ! Test that the target sub-basin is not within the sources (inflows) of the current point
  !
  transfer = .TRUE.
  DO i=1,in_number(sp,sb)
     IF (in_grid(sp,sb,i) == ep .AND. in_basin(sp,sb,i) == eb ) THEN
        WRITE(numout,*) "We have a problem here !", ep, eb, " already flows into ", sp, sb
        transfer = .FALSE.
     ENDIF
  ENDDO
  !
  ! Just in case this was not caught in the calling program.
  !
  IF ( ep <= 0 .OR. ep == undef_int .OR. eb <= 0 .OR. eb == undef_int) THEN
     transfer = .FALSE.
  ENDIF
  !
  IF ( transfer ) THEN
     out_grid(sp,sb) = ep
     out_basin(sp,sb) = eb
     !
     in_number(ep,eb) = in_number(ep,eb) + 1
     IF ( in_number(ep,eb) .LE. basmax ) THEN
        in_grid(ep,eb,in_number(ep,eb)) = sp
        in_basin(ep,eb,in_number(ep,eb)) = sb
     ELSE
        WRITE(numout,*) "routing_updateflow : Too many basins flowing into this point ", ep, eb
        CALL ipslerr_p(3,'routing_updateflow', &
             "Too many basins flowing into this point ",'Increase nbxmax or its derived quantities','')
     ENDIF
  ENDIF
  !
END SUBROUTINE routing_updateflow
!
!! ================================================================================================================================
!! SUBROUTINE 	: routing_sortcoord
!!
!>\BRIEF         This subroutines orders the coordinates to go from North to South and West to East.
!!
!! DESCRIPTION (definitions, functional, design, flags) : None
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S):
!!
!! REFERENCES   : None
!!
!! FLOWCHART    : None
!! \n
!_ ================================================================================================================================

  SUBROUTINE routing_sortcoord(nb_in, coords, direction, nb_out)
    !
    IMPLICIT NONE
    !
!! INPUT VARIABLES
    INTEGER(i_std), INTENT(in)   :: nb_in             !!
    REAL(r_std), INTENT(inout)   :: coords(nb_in)     !!
    !
!! OUTPUT VARIABLES
    INTEGER(i_std), INTENT(out)  :: nb_out            !!
    !
!! LOCAL VARIABLES
    CHARACTER(LEN=2)             :: direction         !!
    INTEGER(i_std)               :: ipos              !!
    REAL(r_std)                  :: coords_tmp(nb_in) !!
    INTEGER(i_std), DIMENSION(1) :: ll                !!
    INTEGER(i_std)               :: ind(nb_in)        !!

!_ ================================================================================================================================
    !
    ipos = 1
    nb_out = nb_in
    !
    ! Compress the coordinates array
    !
    DO WHILE ( ipos < nb_in )
       IF ( coords(ipos+1) /= undef_sechiba) THEN
         IF ( COUNT(coords(ipos:nb_out) == coords(ipos)) > 1 ) THEN
            coords(ipos:nb_out-1) = coords(ipos+1:nb_out) 
            coords(nb_out:nb_in) = undef_sechiba
            nb_out = nb_out - 1
         ELSE
            ipos = ipos + 1
         ENDIF
      ELSE
         EXIT
      ENDIF
    ENDDO
    !
    ! Sort it now
    !
    ! First we get ready and adjust for the periodicity in longitude
    !
    coords_tmp(:) = undef_sechiba
    IF ( INDEX(direction, 'WE') == 1 .OR.  INDEX(direction, 'EW') == 1) THEN
       IF ( MAXVAL(ABS(coords(1:nb_out))) .GT. 160 ) THEN
          coords_tmp(1:nb_out) = MOD(coords(1:nb_out) + 360.0, 360.0)
       ELSE
          coords_tmp(1:nb_out) = coords(1:nb_out)
       ENDIF
    ELSE IF ( INDEX(direction, 'NS') == 1 .OR.  INDEX(direction, 'SN') == 1) THEN
       coords_tmp(1:nb_out) = coords(1:nb_out)
    ELSE
       WRITE(numout,*) 'The chosen direction (', direction,') is not recognized'
       CALL ipslerr_p(3,'routing_sortcoord','The chosen direction is not recognized','First section','')
    ENDIF
    !
    ! Get it sorted out now
    !
    ipos = 1
    !
    IF ( INDEX(direction, 'WE') == 1 .OR. INDEX(direction, 'SN') == 1) THEN
       DO WHILE (COUNT(ABS(coords_tmp(:)-undef_sechiba) > EPSILON(undef_sechiba)*10.) >= 1)
          ll = MINLOC(coords_tmp(:), coords_tmp /= undef_sechiba)
          ind(ipos) = ll(1) 
          coords_tmp(ll(1)) = undef_sechiba
          ipos = ipos + 1
       ENDDO
    ELSE IF ( INDEX(direction, 'EW') == 1 .OR. INDEX(direction, 'NS') == 1) THEN
       DO WHILE (COUNT(ABS(coords_tmp(:)-undef_sechiba) > EPSILON(undef_sechiba)*10.) >= 1)
          ll = MAXLOC(coords_tmp(:), coords_tmp /= undef_sechiba)
          ind(ipos) = ll(1) 
          coords_tmp(ll(1)) = undef_sechiba
          ipos = ipos + 1
       ENDDO
    ELSE
       WRITE(numout,*) 'The chosen direction (', direction,') is not recognized (second)'
       CALL ipslerr_p(3,'routing_sortcoord','The chosen direction is not recognized','Second section','')
    ENDIF
    !
    coords(1:nb_out) = coords(ind(1:nb_out))
    IF (nb_out < nb_in) THEN
       coords(nb_out+1:nb_in) = zero
    ENDIF
    !
  END SUBROUTINE routing_sortcoord
  !
!! ================================================================================================================================
!! SUBROUTINE 	: routing_diagbox
!!
!>\BRIEF         This function will determine if this point is to be diagnosed or not.
!!
!! DESCRIPTION (definitions, functional, design, flags) : None
!!
!! RECENT CHANGE(S): None
!!
!! MAIN OUTPUT VARIABLE(S):
!!
!! REFERENCES   : None
!!
!! FLOWCHART    : None
!! \n
!_ ================================================================================================================================
  !
LOGICAL FUNCTION routing_diagbox(ip, diaglalo)
  !
  INTEGER(i_std), INTENT(in)              :: ip
  REAL(r_std), DIMENSION(:,:), INTENT(in) :: diaglalo
  !
  ! Local
  !
  REAL(r_std) :: dlon, dlat
  INTEGER(i_std) :: ij, nbdiag
  !
  nbdiag=SIZE(diaglalo,DIM=1)
  !
  dlon=ABS(corners(ip,NbSegments,1)-corners(ip,1,1))
  dlat=ABS(corners(ip,NbSegments,2)-corners(ip,1,2))
  DO ij=1,NbSegments-1
     dlon = MAX(dlon,ABS(corners(ip,ij+1,1)-corners(ip,ij,1)))
     dlat = MAX(dlat,ABS(corners(ip,ij+1,2)-corners(ip,ij,2)))
  ENDDO
  !
  routing_diagbox = .FALSE.
  DO ij=1,nbdiag
     IF ( ABS(lalo(ip,1)-diaglalo(ij,1)) < dlat/2.0 .AND. &
        & ABS(lalo(ip,2)-diaglalo(ij,2)) < dlon/2.0 ) THEN
        routing_diagbox = .TRUE.
     ENDIF
  ENDDO
  !
END FUNCTION routing_diagbox
!
LOGICAL FUNCTION routing_diagbox_g(ip, diaglalo)
  !
  INTEGER(i_std), INTENT(in)              :: ip
  REAL(r_std), DIMENSION(:,:), INTENT(in) :: diaglalo
  !
  ! Local
  !
  REAL(r_std) :: dlon, dlat
  INTEGER(i_std) :: ij, nbdiag
  !
  IF (.NOT. is_root_prc) THEN
     WRITE(*,*) "The function routing_diagbox_g is written to work on the root proc and the full grid"
     STOP "routing_diagbox_g"
  ENDIF
  !
  nbdiag=SIZE(diaglalo,DIM=1)
  !
  dlon=ABS(corners_g(ip,NbSegments,1)-corners_g(ip,1,1))
  dlat=ABS(corners_g(ip,NbSegments,2)-corners_g(ip,1,2))
  DO ij=1,NbSegments-1
     dlon = MAX(dlon,ABS(corners_g(ip,ij+1,1)-corners_g(ip,ij,1)))
     dlat = MAX(dlat,ABS(corners_g(ip,ij+1,2)-corners_g(ip,ij,2)))
  ENDDO
  !
  routing_diagbox_g = .FALSE.
  DO ij=1,nbdiag
     IF ( ABS(lalo_g(ip,1)-diaglalo(ij,1)) < dlat/2.0 .AND. &
        & ABS(lalo_g(ip,2)-diaglalo(ij,2)) < dlon/2.0 ) THEN
        routing_diagbox_g = .TRUE.
     ENDIF
  ENDDO
  !
END FUNCTION routing_diagbox_g
!
END MODULE routing_tools
